@@ -90,7 +90,6 @@ __extern int open(const char *, int, ...);
 __extern int openat(int, const char *, int, ...);
 #endif
 __extern int creat(const char *, mode_t);
-__extern int open_cloexec(const char *, int, mode_t);
 __extern int close(int);
 __extern off_t lseek(int, off_t, int);
 /* off_t is 64 bits now even on 32-bit platforms; see llseek.c */
@@ -144,13 +143,13 @@ __extern int optind, opterr, optopt;
 __extern int isatty(int);
 
 __extern unsigned int __page_size;
-static __inline__ int getpagesize(void)
+__must_inline int getpagesize(void)
 {
 	return __page_size;
 }
 
 __extern unsigned int __page_shift;
-static __inline__ int __getpageshift(void)
+__must_inline int __getpageshift(void)
 {
 	return __page_shift;
 }
@@ -162,4 +161,7 @@ __extern int daemon(int, int);
 #define STDOUT_FILENO	1
 #define STDERR_FILENO	2
 
+/* This #include must be at the end */
+#include <sys/sysconf.h>
+
 #endif				/* _UNISTD_H */
