@@ -1,23 +1,32 @@
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <sys/stat.h>
 
 char *progname;
 
 static __noreturn usage(void)
 {
-	fprintf(stderr, "Usage: %s name {b|c|p} major minor\n", progname);
+	fprintf(stderr, "Usage: %s [-m mode] name {b|c|p} major minor\n",
+			progname);
 	exit(1);
 }
 
 int main(int argc, char *argv[])
 {
 	char *name, *type, typec, *endp;
-	unsigned int major, minor;
-	mode_t mode;
+	unsigned int major_num, minor_num;
+	mode_t mode, mode_set = 0;
 	dev_t dev;
 
 	progname = *argv++;
+	if (argc == 1)
+		usage();
+
+	if (argv[0][0] == '-' && argv[0][1] == 'm' && !argv[0][2]) {
+		mode_set = strtoul(argv[1], &endp, 8);
+		argv += 2;
+	}
 
 	name = *argv++;
 	if (!name)
@@ -29,7 +38,7 @@ int main(int argc, char *argv[])
 	typec = type[0];
 
 	mode = 0;
-	switch (type[0]) {
+	switch (typec) {
 	case 'c':
 		mode = S_IFCHR;
 		break;
@@ -49,13 +58,13 @@ int main(int argc, char *argv[])
 		if (!argv[0] || !argv[1])
 			usage();
 
-		major = strtol(*argv++, &endp, 0);
+		major_num = strtol(*argv++, &endp, 0);
 		if (*endp != '\0')
 			usage();
-		minor = strtol(*argv++, &endp, 0);
+		minor_num = strtol(*argv++, &endp, 0);
 		if (*endp != '\0')
 			usage();
-		dev = makedev(major, minor);
+		dev = makedev(major_num, minor_num);
 	}
 
 	if (*argv)
@@ -66,5 +75,10 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
+	if (mode_set && chmod(name, mode_set)) {
+		perror("chmod");
+		exit(1);
+	}
+
 	exit(0);
 }
